# coding=utf-8
# Copyright 2018 Sascha Schirra
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" A ND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import sys
from ropper.common.error import RopperError
try:
    import z3
    from z3 import *
    from z3.z3types import Z3Exception
    import pyvex
    import archinfo
except ImportError as e:
    pass

from ropper.common.utils import toHex, isHex
from ropper.z3helper import create_number_expression, create_register_expression, create_read_memory_expression
import ropper.arch
import ropper.common.enum as enum

import math
import re

if sys.version_info.major == 3:
    long = int

class Category(enum.Enum):
    _enum_ = 'NONE WRITE_MEM_FROM_MEM WRITE_REG_FROM_MEM WRITE_REG_FROM_REG WRITE_MEM_FROM_REG NEG_REG STACK_PIVOTING LOAD_REG LOAD_MEM STACK_PIVOT SYSCALL JMP CALL WRITE_MEM INC_REG CLEAR_REG SUB_REG ADD_REG SUB_REG MUL_REG DIV_REG XCHG_REG PUSHAD WRITE_MEM'

class Analyser(object):

    def __init__(self):
        self.__work = True
        if 'z3' not in globals():
            self.__work = False
            return

    def analyse(self, gadget):
        if not self.__work:
            return False
        try:
            thumb = 1 if isinstance(gadget.arch, ropper.arch.ArchitectureArmThumb) else 0
            irsb = pyvex.IRSB(bytes(gadget.bytes), gadget.address+thumb, gadget.arch.info, bytes_offset=thumb, num_bytes=len(gadget.bytes), opt_level=0)
            irsb_anal = IRSBAnalyser()
            anal = irsb_anal.analyse(irsb)
            archinfo = gadget.arch.info
            anal.spOffset = self.findSpOffset(gadget, anal, archinfo.register_names[archinfo.registers['sp'][0]])
            return anal

        except pyvex.PyVEXError as e:
            pass
        except:
            pass

    def findSpOffset(self, g, anal, sp ):
        if sp not in anal.regs:
            return 0
        slicer = Slicer()
        slice_instructions = []
        slice = slicer.slice(anal.expressions, [sp])
        solver = z3.Solver()

        expr_len = len(anal.expressions)
        expr = None
        tmp = None

        for inst in anal.expressions:

            if not inst:
                continue
            if expr is None:
                expr = inst[0]
            else:
                expr = 'And(%s, %s)' % (expr, inst[0])

        if expr:
            expr = ExpressionBuilder().build(anal.regs, anal.mems, expr)
            sp1 = anal.regs[sp][0]
            sp2 = anal.regs[sp][-1]
            size = int(sp1.split('_')[-1])
            sp1 = z3.BitVec(sp1, size)
            sp2 = z3.BitVec(sp2, size)
            diff = z3.BitVec('diff', size)
            solver.add(z3.And(expr, diff == sp2 - sp1))
            solver.check()
            spOffset = solver.model()[diff].as_signed_long()
            solver = z3.Solver()
            solver.add(z3.And(expr, sp2 - sp1 == spOffset))
            if solver.check() == z3.unsat:
                return 'Undef'
            return spOffset


class InstructionAnalysis(object):

    MEM_COUNTER = 0

    def __init__(self, arch):
        self.__spOffset = 0
        self.__clobberedRegs = {}
        self.__offsets = {}
        self.__tmps = {}
        self.__expressions = []
        self.__usedRegs = set()

    @property
    def usedRegs(self):
        return self.__usedRegs

    def addRegister(self, reg):
        self.__usedRegs.update([reg])

    @property
    def expressions(self):
        return self.__expressions

    @property
    def spOffset(self):
        return self.__spOffset

    @spOffset.setter
    def spOffset(self, offset):
        self.__spOffset = offset

    @property
    def clobberedRegs(self):
        return self.__clobberedRegs

    @property
    def offsets(self):
        return self.__offsets

    @property
    def tmps(self):
        return self.__tmps

    def getValueForTmp(self, tmp):
        while not isinstance(tmp, (int, long)) and tmp is not None:
            tmp = self.tmps.get(tmp)

        return tmp

class Analysis(object):

    def __init__(self, arch, irsb):
        self.__instructions = []
        self.__mem = None
        self.__arch = arch
        self.__registerAccessors = {}
        self.mems = []
        self.__regCount = {}
        self.__mem_counter = 0
        self.irsb = irsb
        self.register_assignments = {}

    def __repr__(self):
        return 'None'

    @property
    def memCounter(self):
        return self.__mem_counter

    @property
    def regs(self):
        return self.__registerAccessors

    @property
    def arch(self):
        return self.__arch

    @property
    def instructions(self):
        return self.__instructions

    @property
    def currentInstruction(self):
        if not len(self.instructions):
            return self.newInstruction()
        return self.__instructions[-1]

    @property
    def clobberedRegs(self):
        to_return = dict()
        for ia in self.instructions:
            to_return.update(ia.clobberedRegs)
        return to_return

    @property
    def usedRegs(self):
        to_return = set()
        for ia in self.instructions:
            to_return.update(ia.usedRegs)
        return to_return

    def newInstruction(self):
        self.__instructions.append(InstructionAnalysis(self.arch))
        return self.__instructions[-1]

    @property
    def spOffset(self):
        offset = 0
        for inst in self.instructions:
            if inst.spOffset is None:
                return None
            offset += inst.spOffset
        return offset

    @property
    def expressions(self):
        to_return = []
        for ia in self.instructions:
            to_return.extend(ia.expressions)
        return to_return

    @property
    def _memory(self):
        if self.__mem is None:
            #self.__mem = z3.Array('memory_%d' % self.__mem_counter , z3.BitVecSort(self.__arch.bits), z3.BitVecSort(8))
            self.__mem = 'memory%d_%d_%d' % (self.__mem_counter, self.__arch.bits, 8)
            self.mems.append(self.__mem)
            self.__mem_counter += 1
        return self.__mem

    def readMemory(self, addr, size, analyse=True):
        return create_read_memory_expression(self._memory, addr, size)

    def writeMemory(self, addr, size, data):
        size = size/8
        old = self._memory
        for i in range(int(size)):
            #old = z3.Store(old, addr+i, z3.Extract((i+1)*8-1,i*8,data))

            value = 'Extract(%d, %d, %s)' % ((i+1)*8-1, i*8, data)
            old = 'Store(%s, %s + %d, %s)' % (old, addr, i, value)

        self.__mem = None
        return '%s == %s' % (old, self._memory)

    def writeRegister(self, offset, size, value):
        reg = self.__arch.translate_register_name(offset & 0xfffffffe, size)

        real_size = self.__arch.registers.get(reg)
        if real_size == None:
            real_size = self.arch.bits*2
        else:
            real_size = real_size[1]*8

        count = self.__regCount.get((reg),0)
        self.__regCount[(reg)] = count + 1

        reg_list = self.__registerAccessors.get((reg))
        if not reg_list:
            reg_list = []
            self.__registerAccessors[(reg)] = reg_list

        self.register_assignments[reg] = value

        reg_list.append('%s_%d_%d' % (reg, count, real_size))

        if size < real_size:
            return 'Extract(%d, 0, %s) == %s' %(size-1,self.__registerAccessors[(reg)][-1],value)
        else:
            return '%s == %s' % (self.__registerAccessors[(reg)][-1], value)

    def __getRegisterAccessor(self, register, size):
        register_list = self.__registerAccessors.get((register))

        if not register_list:
            register_list = ['%s_%d_%d' % (register, self.__regCount.get(register, 0), size)]
            self.__regCount[register] = 1
            self.__registerAccessors[register] = register_list

        return register_list[-1]

    def readRegister(self, offset, size):
        name = offset
        register_size = 0

        if isinstance(name, (int, long)):
            name = self.__arch.translate_register_name(offset & 0xfffffffe, size)
        self.currentInstruction.usedRegs.update([name])
        register_size = self.__arch.registers.get(name)
        if register_size == None:
            register_size = self.arch.bits*2
        else:
            register_size = register_size[1]*8

        reg_acc = self.__getRegisterAccessor(name, register_size)

        return create_register_expression(reg_acc, size, bool(offset & 1) if type(offset) is not str else False)

class SemanticInformation(object):

    def __init__(self, regs, usedRegs, clobberedRegs, mems, expressions, spOffset, checked_constraints=None, irsb=None):

        self.regs = regs
        self.usedRegs = usedRegs
        self.clobberedRegisters = clobberedRegs
        self.mems = mems
        self.expressions = expressions
        self.spOffset = spOffset
        self._checkedConstraints = {} if checked_constraints is None else checked_constraints
        self.irsb = irsb
    @property
    def checkedConstraints(self):
        return self._checkedConstraints

    def __repr__(self):
        return 'SemanticInformation(%s, %s, %s, %s, %s, %s, %s, %s)' % (repr(self.regs), repr(self.usedRegs), repr(self.clobberedRegisters), repr(self.mems), repr(self.expressions), repr(self.spOffset), repr(self.checkedConstraints), repr(self.irsb))

class IRSB_DATA(enum.Enum):
    _enum_ = 'WRITE_REG READ_REG SP_OFFSET CONSTANT'


class CommandClass(object):

    @classmethod
    def use(cls, name):
        if not isinstance(name, str):
            name = name.__class__.__name__.lower()
        return getattr(cls, name, cls.dummy)

    @staticmethod
    def dummy(*args, **kwargs):
        pass


class ZExpressions(CommandClass):

    @staticmethod
    def get(dest, data, analysis):
        return (analysis.readRegister(data.offset, data.result_size(analysis.irsb.tyenv)),(analysis.arch.translate_register_name(data.offset & 0xfffffffe, data.result_size(analysis.irsb.tyenv)),))

    @staticmethod
    def load(dest, data, analysis):
        addr = ZExpressions.use(data.addr)(dest, data.addr, analysis)[0]
        return (analysis.readMemory(addr, data.result_size(analysis.irsb.tyenv)),(addr,))

    @staticmethod
    def store(dest, data, analysis):
        addr = ZExpressions.use(data.addr)(dest, data.addr, analysis)[0]
        return (analysis.readMemory(addr, data.result_size(analysis.irsb.tyenv), False),(addr,))

    @staticmethod
    def const(dest, data, analysis):
        analysis.currentInstruction.tmps[dest] = data.con.value if not math.isnan(data.con.value) else 0
        return ('BitVecVal(%d, %d)' % (analysis.currentInstruction.tmps[dest], data.con.size), (data.con.value,))

    @staticmethod
    def rdtmp(dest, data, analysis):
        tmp = '%s_%d' % (str(data), data.result_size(analysis.irsb.tyenv))
        analysis.currentInstruction.tmps[dest] = tmp
        analysis.regs[tmp] = [tmp]
        return ('%s' % (tmp), (tmp,))

    @staticmethod
    def binop(dest, data, analysis):
        arg1 = ZExpressions.use(data.args[0])(dest, data.args[0], analysis)
        arg2 = ZExpressions.use(data.args[1])(dest, data.args[1], analysis)

        return (ZOperations.use(data.op)(arg1[0], arg2[0], analysis), (arg1[0], arg2[0]))

    @staticmethod
    def unop(dest, data, analysis):
        arg1 = ZExpressions.use(data.args[0])(dest, data.args[0], analysis)
        return (ZOperations.use(data.op)(arg1[0], analysis), arg1[1])

    @staticmethod
    def triop(dest, data, analysis):
        # TODO used for floating point operations, should be implemented in a future version
        return None

    @staticmethod
    def qop(dest, data, analysis):
        # TODO should be implemented in a future version
        return None

    @staticmethod
    def ccall(dest, data, analysis):
        # TODO should be implemented in a future version
        #analysis.printable=True
        return None

    @staticmethod
    def dummy(dest, data, analysis):
        pass


class ZOperations(CommandClass):

    @staticmethod
    def Iop_Add32(arg1, arg2, analysis):
        return '%s + %s' % (arg1, arg2)

    @staticmethod
    def Iop_Add16(arg1, arg2, analysis):
        return '%s + %s' % (arg1, arg2)

    @staticmethod
    def Iop_Add8(arg1, arg2, analysis):
        return '%s + %s' % (arg1, arg2)

    @staticmethod
    def Iop_Xor32(arg1, arg2, analysis):
        return '%s ^ %s' % (arg1, arg2)

    @staticmethod
    def Iop_Xor16(arg1, arg2, analysis):
        return '%s ^ %s' % (arg1, arg2)

    @staticmethod
    def Iop_Xor8(arg1, arg2, analysis):
        return '%s ^ %s' % (arg1, arg2)

    @staticmethod
    def Iop_Mul32(arg1, arg2, analysis):
        return '%s * %s' % (arg1, arg2)

    @staticmethod
    def Iop_Mul16(arg1, arg2, analysis):
        return '%s * %s' % (arg1, arg2)

    @staticmethod
    def Iop_Mul8(arg1, arg2, analysis):
        return '%s * %s' % (arg1, arg2)

    @staticmethod
    def Iop_Div32(arg1, arg2, analysis):
        return '%s / %s' % (arg1, arg2)

    @staticmethod
    def Iop_Div16(arg1, arg2, analysis):
        return '%s / %s' % (arg1, arg2)

    @staticmethod
    def Iop_Div8(arg1, arg2, analysis):
        return '%s / %s' % (arg1, arg2)

    @staticmethod
    def Iop_Sub32(arg1, arg2, analysis):
        return '%s - %s' % (arg1, arg2)

    @staticmethod
    def Iop_Sub16(arg1, arg2, analysis):
        return '%s - %s' % (arg1, arg2)

    @staticmethod
    def Iop_Sub8(arg1, arg2, analysis):
        return '%s - %s' % (arg1, arg2)

    @staticmethod
    def Iop_Add64(arg1, arg2, analysis):
        return '%s + %s' % (arg1, arg2)

    @staticmethod
    def Iop_Xor64(arg1, arg2, analysis):
        return '%s ^ %s' % (arg1, arg2)

    @staticmethod
    def Iop_And64(arg1, arg2, analysis):
        return '%s & %s' % (arg1, arg2)

    @staticmethod
    def Iop_And32(arg1, arg2, analysis):
        return '%s & %s' % (arg1, arg2)

    @staticmethod
    def Iop_And16(arg1, arg2, analysis):
        return '%s & %s' % (arg1, arg2)

    @staticmethod
    def Iop_And8(arg1, arg2, analysis):
        return '%s & %s' % (arg1, arg2)

    @staticmethod
    def Iop_Or64(arg1, arg2, analysis):
        return '%s | %s' % (arg1, arg2)

    @staticmethod
    def Iop_Or32(arg1, arg2, analysis):
        return '%s | %s' % (arg1, arg2)

    @staticmethod
    def Iop_Or16(arg1, arg2, analysis):
        return '%s | %s' % (arg1, arg2)

    @staticmethod
    def Iop_Or8(arg1, arg2, analysis):
        return '%s | %s' % (arg1, arg2)


    @staticmethod
    def Iop_Mul64(arg1, arg2, analysis):
        return '%s * %s' % (arg1, arg2)

    @staticmethod
    def Iop_Div64(arg1, arg2, analysis):
        return '%s / %s' % (arg1, arg2)

    @staticmethod
    def Iop_Sub64(arg1, arg2, analysis):
        return '%s - %s' % (arg1, arg2)

    @staticmethod
    def Iop_32Uto64(arg1, analysis):
        return 'ZeroExt(32,%s)' % arg1

    @staticmethod
    def Iop_32to64(arg1, analysis):
        return 'SignExt(32,%s)' % arg1

    @staticmethod
    def Iop_8to32(arg1, analysis):
        return 'SignExt(24,%s)' % arg1

    @staticmethod
    def Iop_16to32(arg1, analysis):
        return 'SignExt(16,%s)' % arg1

    @staticmethod
    def Iop_8to32(arg1, analysis):
        return 'ZeroExt(24,%s)' % arg1

    @staticmethod
    def Iop_8Uto32(arg1, analysis):
        return 'ZeroExt(24,%s)' % arg1

    @staticmethod
    def Iop_16Uto32(arg1, analysis):
        return 'ZeroExt(16,%s)' % arg1

    @staticmethod
    def Iop_64Uto32(arg1, analysis):
        return 'Extract(31,0,%s)' % arg1

    @staticmethod
    def Iop_64to32(arg1, analysis):
        return 'Extract(31,0,%s)' % arg1

    @staticmethod
    def Iop_32to8(arg1, analysis):
        return 'Extract(7,0,%s)' % arg1

    @staticmethod
    def Iop_32Uto8(arg1, analysis):
        return 'Extract(7,0,%s)' % arg1

    @staticmethod
    def Iop_32to16(arg1, analysis):
        return 'Extract(15,0,%s)' % arg1

    @staticmethod
    def Iop_32Uto16(arg1, analysis):
        return 'Extract(15,0,%s)' % arg1


class ZStatements(CommandClass):

    @staticmethod
    def put(stmt, analysis):
        dest = analysis.arch.translate_register_name(stmt.offset, stmt.data.result_size(analysis.irsb.tyenv))
        value = ZExpressions.use(stmt.data)(dest,stmt.data, analysis)

        if not dest.startswith('cc_'):
            analysis.currentInstruction.clobberedRegs[dest] = None

        if stmt.offset == analysis.arch.sp_offset:
            analysis.currentInstruction.spOffset = analysis.currentInstruction.getValueForTmp(str(stmt.data))

        if value is not None:
            analysis.register_assignments[dest] = value[1][0] if len(value[1]) == 1 else value[0]
        else:
            analysis.register_assignments[dest] = None


        return (analysis.writeRegister(stmt.offset, stmt.data.result_size(analysis.irsb.tyenv), value[0]),dest, value[1])

    @staticmethod
    def wrtmp(stmt, analysis):
        tmp = 't'+str(stmt.tmp)
        value = ZExpressions.use(stmt.data)( tmp, stmt.data, analysis)
        tmp = '%s_%s' % (tmp, stmt.data.result_size(analysis.irsb.tyenv))
        analysis.regs[tmp] = [tmp]
        if value is not None:
            analysis.register_assignments[tmp] = value[1][0] if len(value[1]) == 1 else value[0]
        else:
            analysis.register_assignments[tmp] = None
        if value is None or value[0] is None:
            return False

        return ('%s == %s' % (tmp, value[0]), tmp, value[1])

    @staticmethod
    def store(stmt, analysis):
        addr = ZExpressions.use(stmt.addr)(None, stmt.addr, analysis)[0]
        value = ZExpressions.use(stmt.data)(str(addr), stmt.data, analysis)

        return (analysis.writeMemory(addr, stmt.data.result_size(analysis.irsb.tyenv), value[0]),addr[0],value[1])

    @staticmethod
    def imark(stmt, analysis):
        analysis.newInstruction()

    @staticmethod
    def dummy(stmt, analysis):
        pass

class IRSBAnalyser(object):

    def __init__(self):
        self.__cRegs = []

    def __resolveAssignments(self, analysis):
        to_return = {}
        for orig in analysis.clobberedRegs.keys():
            reg = orig
            while True:
                tmp = analysis.register_assignments.get(reg)
                if tmp is None:
                    to_return[orig] = None
                    break
                elif analysis.arch.registers.get(tmp) is not None:
                    to_return[orig] = tmp
                    break

                reg = tmp

        return to_return


    def analyse(self, irsb):
        anal = Analysis(irsb.arch, irsb)
        sp_offset = 0
        for stmt in irsb.statements:
            name = stmt.__class__.__name__.lower()
            func = ZStatements.use(stmt)
            anal.printable = False
            expr = func(stmt, anal)
           # if anal.printable:
            #    print(stmt)
            ci = anal.currentInstruction
            ci.expressions.append(expr)

        clobbered_regs = self.__resolveAssignments(anal)
        return SemanticInformation(anal.regs, anal.usedRegs, clobbered_regs, anal.mems, anal.expressions, anal.spOffset, irsb=irsb._pp_str())


class Slice(object):

    def __init__(self, regs):
        self.instructions = []
        self.expressions = []
        self.regs = []
        self.regs.extend(regs)


class Slicer(object):

    def slice(self, irsb, reg):
        slice = Slice(reg)

        for expr in irsb[::-1]:
            if expr and expr[1] in slice.regs:
                if expr[2][0] and type(expr[2][0]) is str and not expr[2][0].isdigit():
                    slice.regs.append(expr[2][0])
                if len(expr[2]) == 2 and type(expr[2][1]) is str and not expr[2][1].isdigit():
                    slice.regs.append(expr[2][1])
                if expr[0]:
                    slice.expressions.append(expr[0])

        return slice


class ExpressionBuilder(object):

    def __init__(self):
        self.__z3objects = {}

    def __getattr__(self, name):
        if name in self.__z3objects:
            return self.__z3objects[name]
        return super(ExpressionBuilder, self).__getattribute__(name)

    def _createRegs(self, objects, regsDict):
        for regs in regsDict.values():
            for reg in regs:
                objects[reg] = z3.BitVec(reg, int(reg.split('_')[-1],10))

    def _createMem(self, objects, mems):
        for mem in mems:
            sizes = mem.split('_')
            objects[mem] = z3.Array(mem, z3.BitVecSort(int(sizes[-2],10)), z3.BitVecSort(int(sizes[-1],10)))

    def build(self, regs, mems, expression, constraint=None):
        z3objects = {}
        self._createRegs(z3objects, regs)
        self._createMem(z3objects, mems)
        if constraint is None:
            return eval(expression, globals(), z3objects)
        f = True
        f = z3.And(f,eval(expression, globals(), z3objects))
        g = eval(constraint, globals(), z3objects)

        return z3.And(f, z3.Not(g))
#        return z3.Not(eval(expression, globals(), z3objects) == eval(constraint, globals(), z3objects))
#        return z3.And(eval(expression, globals(), z3objects), z3.Not(eval(constraint, globals(), z3objects))



# coding=utf-8
# Copyright 2018 Sascha Schirra
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" A ND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ropper.loaders.loader import *
from ropper.rop import Ropper
from ropper.arch import *
from ropper.gadget import Gadget


import unittest

# class ELF_ARM(unittest.TestCase):
#
#     def setUp(self):
#         self.file = Loader.open('test-binaries/ls-arm')
#         self.file.arch = ARM
#
#     def test_general(self):
#         self.assertEqual(self.file.arch, ARM)
#         self.assertEqual(self.file.type, Type.ELF)
#
#
#     def test_gadgets(self):
#         ropper = Ropper()
#         gadgets = ropper.searchGadgets(self.file)
#
#         gadget = gadgets[0]
#         self.assertEqual(len(gadgets), 1711)
#         self.assertEqual(gadget.lines[0][0], 0x8567)
#         self.assertEqual(gadget.imageBase, 0x8048000)
#         self.file.imageBase = 0x0
#         self.assertEqual(gadget.imageBase, 0x0)
#         self.file.imageBase = None
#         self.assertEqual(gadget.imageBase, 0x8048000)




class ELF_ARM_THUMB(unittest.TestCase):

    def setUp(self):
        self.file = Loader.open('test-binaries/ls-arm')


    def test_general(self):
        self.assertEqual(self.file.arch, ARMTHUMB)
        self.assertEqual(self.file.type, Type.ELF)

    def test_gadgets_pe(self):
        ropper = Ropper()
        gadgets = ropper.searchGadgets(self.file)

        gadget = gadgets[0]
        self.assertGreater(len(gadgets), 1300)
        self.assertEqual(gadget.lines[0][0] + self.file.imageBase, gadget.address)
        self.assertEqual(gadget.imageBase, 0x00008000)
        self.file.imageBase = 0x0
        Gadget.IMAGE_BASES[self.file.checksum] = self.file.imageBase
        self.assertEqual(gadget.imageBase, 0x0)
        self.file.imageBase = None
        Gadget.IMAGE_BASES[self.file.checksum] = self.file.imageBase
        self.assertEqual(gadget.imageBase, 0x00008000)


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
# Copyright 2018 Sascha Schirra
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" A ND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ropper.loaders.loader import *
from ropper.rop import Ropper
from ropper.arch import *
from ropper.common.error import *
from ropper.service import RopperService

from filebytes.binary import BinaryError

from sys import version_info
import unittest
import os
import ropper

FILE = 'test-binaries/ls-x86_64'

class GeneralTests(unittest.TestCase):

    def setUp(self):
        self.rs = RopperService()
        self.rs.addFile('test-binaries/ls-x86_64')
        self.rs.loadGadgetsFor()

    def test_search(self):

        found_gadgets = self.rs.searchdict('mov [rax]')[FILE]
        self.assertEqual(len(found_gadgets), 2)

        found_gadgets = self.rs.searchdict('mov [r?x%]')[FILE]
        self.assertGreater(len(found_gadgets), 12)


    def test_badbytes(self):

        self.rs.options.badbytes = 'adfd'

        badbytes = 'adfd'

        gadget = self.rs.files[0].gadgets[0]
        self.assertNotEqual(gadget.lines[0][0], 0x1adfd)


        self.rs.options.badbytes =  'b1c7'
        gadgets = self.rs.searchJmpReg(['rsp'])
        self.assertNotEqual(gadgets[FILE][0].lines[0][0], 0xb1c7)

        with self.assertRaises(AttributeError):
            self.rs.options.badbytes = 'b1c'


        with self.assertRaises(AttributeError):
            self.rs.options.badbytes = 'qwer'


    def test_opcode_failures(self):
        r = RopperService()

        if version_info.major == 3 and version_info.minor >= 2:
            # Wrong question mark position
            with self.assertRaisesRegex(RopperError,'A \? for the highest 4 bit of a byte is not supported.*'):
                self.rs.searchOpcode('ff?4')
            # Wrong lengh
            with self.assertRaisesRegex(RopperError,'The length of the opcode has to be a multiple of two'):
                self.rs.searchOpcode('ff4')
            # Unallowed character
            with self.assertRaisesRegex(RopperError,'Invalid characters in opcode string'):
                self.rs.searchOpcode('ff4r')
        else:
            # Wrong question mark position
            with self.assertRaisesRegexp(RopperError,'A \? for the highest 4 bit of a byte is not supported.*'):
                self.rs.searchOpcode('ff?4')
            # Wrong lengh
            with self.assertRaisesRegexp(RopperError,'The length of the opcode has to be a multiple of two'):
                self.rs.searchOpcode('ff4')
            # Unallowed character
            with self.assertRaisesRegexp(RopperError,'Invalid characters in opcode string'):
                self.rs.searchOpcode('ff4r')


class RegressionTests(unittest.TestCase):
    def test_segfault_pe_001(self):
        with self.assertRaises(BinaryError):
            _ = Loader.open('test-binaries/hang-mutate_bytes2_3da2c4818ffe26a52b06b348969026f3_.exe')



# coding=utf-8
# Copyright 2018 Sascha Schirra
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" A ND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
from ropper.common.error import RopperError

class ConstraintCompiler(object):
    """
    Compile a user given constraints to z3 expressions

    constraint := assignment | pop_reg
    assignment := reg, adjust, reg | number
    pop_reg := "pop", reg
    adjust := "==" | "+=" | "-=" | "*=" | "/="
    reg := a register of the current architecture
    number := int
    """

    NUMBER_REGEX = '(-?[0-9]+)'
    REG_REGEX = '(?P<{}>[a-zA-Z0-9]+)'
    ADJUST_REGEX = '([\\+\-\*/=]=)'
    ASSIGNMENT_REGEX = '('+REG_REGEX.format('reg_dst_1') + ' *' + ADJUST_REGEX + ' *('+NUMBER_REGEX+'|'+REG_REGEX.format('reg_src_1')+'|(\[)'+REG_REGEX.format('reg_src_2')+'(\])))'
    POP_REGEX = '((pop) +'+REG_REGEX.format('reg_dst_2')+')'
    CONSTRAINT_REGEX = '(' + ASSIGNMENT_REGEX + '|' + POP_REGEX + ')'

    def __init__(self, architecture, semantic_info):
        self.__architecture = architecture
        self.__semantic_info = semantic_info

    def getSymbols(self, constraints):
        symbols = []
        for constraint in constraints:
            match = re.match(ConstraintCompiler.CONSTRAINT_REGEX, constraint)
            if match is None:
                raise Exception('Invalid syntax: %s' % constraint)
            reg_dst = match.group('reg_dst_1')
            if reg_dst is not None:
                reg_src = match.group('reg_src_1')
                reg_src = match.group('reg_src_2') if reg_src is None else reg_src
                symbols.append((reg_dst, reg_src))
            else:
                symbols.append((match.group('reg_dst_2'), None))

        return symbols


    def compile(self, constraints):
        """
        compile a line of semantic expressions
        """
        tokens = self._tokenize(constraints)[::-1]
        to_return = None
        constraint = None
        while True:
            if not tokens:
                break

            token = tokens.pop()
            if token in self.__architecture.info.registers:
                constraint = self._assignment(token, tokens)
            elif token == 'pop':
                constraint = self._popReg(token, tokens)
            elif token == ';':
                if to_return is None:
                    to_return = constraint
                else:
                    to_return = 'And(%s, %s)' % (to_return, constraint)
            else:
                raise ConstraintError('Invalid token: %s' % token)

        return to_return

    def _tokenize(self, constraints):
        """
        return a list of tokens
        """
        tokens = []
        for constraint in constraints.split(';'):
            constraint = constraint.strip()
            if not constraint:
                continue
            match = re.match(ConstraintCompiler.CONSTRAINT_REGEX, constraint)
            if match is None:
                raise ConstraintError('Invalid Syntax: %s' % constraint)
            last_valid_index = -1
            for index in range(1, len(match.regs)):
                start = match.regs[index][0]
                if start == -1:
                    continue
                if last_valid_index == -1:
                    last_valid_index = index
                    continue
                if match.regs[last_valid_index][0] != start:
                    tokens.append(match.group(last_valid_index))
                last_valid_index = index
            tokens.append(match.group(last_valid_index))
            tokens.append(';')
        return tokens

    def _assignment(self, register, tokens):
        register = self.__architecture.getRegisterName(register)
        reg1_last = self.__semantic_info.regs[register][-1]
        reg1_init = self.__semantic_info.regs[register][0]
        op = tokens.pop()
        if not re.match(ConstraintCompiler.ADJUST_REGEX, op):
            raise ConstraintError('Invalid syntax: %s' % op)
        value = tokens.pop()
        if value == '[':
            r1 = register
            register = tokens.pop()

            register_name = self.__architecture.getRegisterName(register)
            if not register_name:
                raise ConstraintError('Invalid register: %s' & register)
            value = self._readMemory(register_name)
            tokens.pop()
        elif re.match(ConstraintCompiler.NUMBER_REGEX, value):
            value = create_number_expression(int(value), int(reg1_last.split('_')[-1]))

        elif value in self.__architecture.info.registers:
            value = self.__architecture.getRegisterName(value)
            value = self.__semantic_info.regs[value][0]
            value = create_register_expression(value, int(value.split('_')[-1]))
        else:
            print(re.match(ConstraintCompiler.NUMBER_REGEX, value))
            raise ConstraintError('Invalid Assignment: %s%s%s' % (register, op, value))
        reg1_last = create_register_expression(reg1_last, int(reg1_last.split('_')[-1]))
        reg1_init = create_register_expression(reg1_init, int(reg1_init.split('_')[-1]))
        return self._create(reg1_last, reg1_init, value, op[0])

    def _create(self, left_last, left_init, right, adjust):
        if adjust != '=':
            return '%s == %s %s %s' % (left_last, left_init, adjust, right)
        else:
            return '%s == %s' % (left_last, right)

    def _readMemory(self, register):
        register_init = self.__semantic_info.regs[register][0]
        if self.__semantic_info.mems:
            memory = self.__semantic_info.mems[-1]
        else:
            memory = 'memory%d_%d_%d' % (0, self.__architecture.info.bits, 8)
            self.__semantic_info.mems.append(memory)
        size = int(register_init.split('_')[-1])
        register_expr = create_register_expression(register_init, size)
        mem_expr = create_read_memory_expression(memory, register_expr, size)
        return mem_expr

    def _popReg(self, pop, tokens):
        reg_name = tokens.pop()
        self.symbols.append((reg_name,None))
        reg = self.__semantic_info.regs[reg_name][-1]
        if self.__semantic_info.mems:
            memory = self.__semantic_info.mems[0]
        else:
            memory = 'memory%d_%d_%d' % (0, self.__architecture.info.bits, 8)
            self.__semantic_info.mems.append(memory)
        size = int(reg.split('_')[-1])
        register_expr = create_register_expression(reg, size)
        mem_expr = create_read_memory_expression(memory, register_expr, size)
        return mem_expr


class ConstraintError(RopperError):
    """
    ConstraintError
    """
    pass


def create_register_expression(register_accessor, size, high=False):
    register_size = int(register_accessor.split('_')[2])
    if size < register_size:
        if high:
            return 'Extract(%d, 8, %s)' % (size+8-1, register_accessor)
        else:
            return 'Extract(%d, 0, %s)' % (size-1, register_accessor)
    else:
        return '%s' % register_accessor

def create_number_expression(number, size):
    return "BitVecVal(%d, %d)" % (number, size)

def create_read_memory_expression(memory, addr, size):
    to_return = '%s[%s]' % (memory, addr)
    for i in range(1, int(size/8)):
        value = '%s[%s]' % (memory, '%s + %d' % (addr, i))
        to_return = 'Concat(%s, %s)' % (value, to_return)

    return to_return
